import z3
from collections import Counter
from typing import NamedTuple, Optional, List, Tuple, Dict, cast
from rfb_mc.component.helper.z3_helper import clone_expression, deserialize_expression, \
    serialize_expression
from rfb_mc.runner import RunnerBase
from rfb_mc.types import Params, RfBmcTask, RfBmcResult, BmcResult, BmcTask

FormulaParamsZ3 = NamedTuple("FormulaParamsZ3", [("formula", z3.BoolRef), ("variables", List[z3.BitVecRef])])

RfmiGenerationArgsZ3 = NamedTuple("RfmiGenerationArgsZ3", [("variables", List[z3.BitVecRef])])


class RunnerZ3(RunnerBase[FormulaParamsZ3, RfmiGenerationArgsZ3, z3.BoolRef]):
    def __init__(
        self,
        params: Params,
        formula_params: FormulaParamsZ3,
    ):
        super().__init__(params, FormulaParamsZ3(
            # translates the contexts, thus the formula and variables need not be created with the context
            # the z3 runner will use
            formula=formula_params.formula.translate(z3.main_ctx())
            if formula_params.formula.ctx != z3.main_ctx() else formula_params.formula,
            variables=[
                x.translate(z3.main_ctx()) if x.ctx != z3.main_ctx() else x
                for x in formula_params.variables
            ],
        ))

        # maps q to a solver that has a q-times conjunction asserted
        self._solver_map: Dict[int, Tuple[z3.Solver, List[z3.BitVecRef]]] = {}

    def _get_solver(self, q: int) -> (z3.Solver, List[z3.BitVecRef]):
        """
        Returns the solver and cloned variables, of which the solver
        has the q-times conjunction of the formula asserted.
        """

        if q not in self._solver_map:
            formula_clone = clone_expression(self.formula_params.formula, q, self.formula_params.variables)

            variables = [
                cast(z3.BitVecRef, q_bv)
                for bv in self.formula_params.variables
                for q_bv in formula_clone.var_map[bv]
            ]

            # TODO: investigate performance of different configurations
            #  and allow for custom solver configurations
            solver = z3.Solver()
            solver.add(z3.And(formula_clone.clones))

            self._solver_map[q] = (solver, variables)

        return self._solver_map[q]

    @classmethod
    def check_params_and_formula_params_compatibility(
        cls,
        params: Params,
        formula_params: FormulaParamsZ3,
    ):
        formula_variable_counter = Counter([x.size() for x in formula_params.variables])
        assert formula_variable_counter == params.bit_width_counter, \
               "bit widths of the formula params must match bit widths of the params"

    @staticmethod
    def _z3_bounded_model_count(solver: z3.Solver, variables: List[z3.ExprRef], u: int) -> Optional[int]:
        """
        If the solver assertions have less than u models that are distinct for the given variables it
        returns the exact model count, otherwise it returns None.
        :param solver:
        :param variables:
        :param u:
        """

        solver.push()

        for i in range(u):
            response = solver.check()

            if response == z3.unknown:
                solver.pop()
                raise RuntimeError("Solver responded with unknown")
            elif response == z3.unsat:
                solver.pop()
                return i

            # in the last iteration adding the constraint would be unnecessary, thus is skipped
            if i != u-1:
                # add assertion that found model cannot be satisfying again
                m = solver.model()
                solver.add(z3.Or([x != m[x] for x in variables]))

        solver.pop()

        return None

    def get_restrictive_formula_instance_generation_args(self, q: int) -> RfmiGenerationArgsZ3:
        _, variables = self._get_solver(q)

        return RfmiGenerationArgsZ3(
            variables=variables,
        )

    def rf_bmc(self, task: RfBmcTask) -> RfBmcResult:
        solver, variables = self._get_solver(task.q)

        # TODO: ensure solver stack is maintained correctly even on error

        solver.push()
        solver.add(
            self.generate_restrictive_formula_instance(task.rfm_guid, task.rfm_formula_params, task.q)
        )

        # is None if solver has at least a models for q_bits
        bmc = self._z3_bounded_model_count(solver, variables, task.a)

        solver.pop()

        return RfBmcResult(bmc=bmc)

    def bmc(self, task: BmcTask) -> BmcResult:
        solver, variables = self._get_solver(1)

        # is None if solver has at least a models
        bmc = self._z3_bounded_model_count(solver, variables, task.a)

        return BmcResult(bmc=bmc)


SerializedFormulaParamsZ3 = NamedTuple(
    "SerializedFormulaParamsZ3",
    [("serialized_formula", str), ("serialized_variables", List[Tuple[str, int]])],
)


def serialize_formula_params_z3(formula_params: FormulaParamsZ3) -> SerializedFormulaParamsZ3:
    return SerializedFormulaParamsZ3(
        serialized_formula=serialize_expression(formula_params.formula),
        serialized_variables=[(str(x), x.size()) for x in formula_params.variables],
    )


def deserialize_formula_params_z3(
    serialized_formula_params: SerializedFormulaParamsZ3,
) -> FormulaParamsZ3:
    return FormulaParamsZ3(
        formula=cast(z3.BoolRef, deserialize_expression(serialized_formula_params.serialized_formula)),
        variables=[z3.BitVec(name, k) for name, k in serialized_formula_params.serialized_variables],
    )



from dataclasses import dataclass
from typing import NamedTuple, Optional, Counter, Any

BmcTask = NamedTuple("BmcTask", [
    ("a", int),
])
""" Parameters for a bounded model counting call """

BmcResult = NamedTuple("BmcResult", [("bmc", Optional[int])])
""" Result of a bounded model counting call """


RfBmcTask = NamedTuple("RfBmcTask", [
    ("rfm_guid", str), ("rfm_formula_params", Any), ("a", int), ("q", int)
])
""" Parameters for a restrictive formula bounded model counting call """


RfBmcResult = NamedTuple("RfBmcResult", [("bmc", Optional[int])])
""" Result of a restrictive formula bounded model counting call """


@dataclass
class Params:
    # keys represent bit widths and values/counts the amount of variable with the given bit width
    bit_width_counter: Counter[int]

    def __post_init__(self):
        if any([bw <= 0 for bw in self.bit_width_counter.keys()]):
            raise ValueError("bit_width_counter cannot contain bit widths (keys) that are <= 0")

        if any([n < 0 for n in self.bit_width_counter.values()]):
            raise ValueError("bit_width_counter cannot contain negative counts")



from abc import abstractmethod
from fractions import Fraction
from math import sqrt, prod, ceil, floor
from typing import Tuple, Optional, Union, Generic, TypeVar, List, Counter
from rfb_mc.component.eamp.eamp_rfm import EampParams, EampRfm
from rfb_mc.component.eamp.types import ProbabilisticInterval
from rfb_mc.component.eamp.utility import majority_vote_error_probability, \
    multi_majority_vote_iteration_count_to_ensure_beta, probability_of_correctness
from rfb_mc.scheduler import SchedulerBase
from rfb_mc.store import StoreBase
from rfb_mc.types import RfBmcTask, RfBmcResult, BmcTask, BmcResult

PartialEampParams = TypeVar("PartialEampParams")


class EampEdgeSchedulerBase(
    Generic[PartialEampParams],
    SchedulerBase[ProbabilisticInterval, ProbabilisticInterval]
):
    def __init__(
        self,
        store: StoreBase,
        confidence: Union[Fraction, float],
        a: int,
        q: int,
        min_model_count: Optional[int] = None,
        max_model_count: Optional[int] = None,
    ):

        super().__init__(store)

        # amount of models that are at most possible for a formula having the amount of bits specified
        theoretical_max_model_count = prod([
            2 ** (bit_width * amount) for bit_width, amount in store.data.params.bit_width_counter.items()
        ])

        assert 0 <= confidence < 1, "Parameter 'confidence' is at least 0 and less than 1"
        assert 1 <= a, "Parameter 'a' is at least 1"
        assert 1 <= q, "Parameter 'q' is at least 1"
        assert min_model_count is None or 0 <= min_model_count, "Parameter 'min_model_count' is at least 0"
        assert min_model_count is None or max_model_count is None or min_model_count <= max_model_count, \
            "Parameter 'min_model_count' is less than or equal parameter 'max_model_count'"
        assert min_model_count is None or min_model_count <= theoretical_max_model_count, \
            "Parameter 'min_model_count' is less than or equal the theoretical maximal model count of the formula," \
            " i.e. 2**(amount of bits in variables)"

        self.min_model_count: int = min_model_count if min_model_count is not None else 0
        self.max_model_count: int = \
            min(max_model_count, theoretical_max_model_count) if max_model_count is not None \
            else theoretical_max_model_count

        self.a: int = a
        self.q: int = q
        self.confidence: Fraction = Fraction(confidence)
        self.store: StoreBase = store

        g, lg = EampEdgeSchedulerBase.get_g_and_lg(a)
        self.g: float = g
        self.lg: float = lg

    @abstractmethod
    def _get_upper_bound_on_estimate_iteration_count(self) -> int:
        """
        Returns an upper bound on the necessary amount of majority vote counting estimate calls that
        will need to be performed.
        """
        
        raise NotImplementedError()

    @abstractmethod
    def _get_required_minimal_min_model_count_when_no_lower_bound_could_be_established(self) -> int:
        """
        Returns the minimal min_model_count required to ensure that, if estimate never responded with True,
        the lower bound, established by min_model_count, will satisfy the multiplicative gap constraints on
        the final interval.
        """

        raise NotImplementedError()

    @abstractmethod
    def _make_eamp_params(self, partial_eamp_params: PartialEampParams) -> EampParams:
        raise NotImplementedError()

    @abstractmethod
    def _make_initial_partial_eamp_params(self) -> PartialEampParams:
        """
        Returns the initial partial eamp parameters with which the algorithm will
        start the estimate iterations.
        """

        raise NotImplementedError()

    @abstractmethod
    def _advance_partial_eamp_params(
        self,
        partial_eamp_params: PartialEampParams,
        estimate_result: bool
    ) -> Optional[PartialEampParams]:
        """
        Returns the next partial eamp parameters as a response to the estimate result.
        If the return value is None, it is assumed the iteration procedure has finished.
        """

        raise NotImplementedError()

    def _run_algorithm_once(self):
        if self.min_model_count == self.max_model_count:
            return ProbabilisticInterval(
                lower_bound=self.min_model_count,
                upper_bound=self.max_model_count,
                confidence=Fraction(1)
            )

        g, lg = EampEdgeSchedulerBase.get_g_and_lg(self.a)
        
        mv_estimate_count_upper_bound = self._get_upper_bound_on_estimate_iteration_count()

        # maximal allowed error probability of the algorithm
        beta = 1 - self.confidence

        r = multi_majority_vote_iteration_count_to_ensure_beta(
            Fraction(1, 4),
            beta,
            mv_estimate_count_upper_bound,
        )

        def make_rf_bmc_task(eamp_params: EampParams):
            return RfBmcTask(
                rfm_guid=EampRfm.get_guid(),
                rfm_formula_params=eamp_params,
                a=self.a,
                q=self.q,
            )

        def range_size(p: PartialEampParams):
            return EampRfm.get_restrictive_formula_properties(
                self.store.data.params, self._make_eamp_params(p),
            ).range_size

        def pre_estimate(p: PartialEampParams) -> Optional[bool]:
            if self.max_model_count ** self.q < range_size(p) * lg:
                return False
            elif self.min_model_count ** self.q > range_size(p) * g:
                return True
            elif p_neg is not None and range_size(p_neg) <= range_size(p):
                return False
            elif p_pos is not None and range_size(p) <= range_size(p_pos):
                return True
            else:
                return None

        p_pos: Optional[PartialEampParams] = None

        p_neg: Optional[PartialEampParams] = None

        # error probability of the independent probabilistic execution that have occurred
        error_probabilities: List[Fraction] = []

        min_model_count = self.min_model_count
        max_model_count = self.max_model_count

        def get_edge_interval():
            if p_pos is not None:
                lower_bound = max(int(ceil((range_size(p_pos) * g) ** (1 / self.q))), min_model_count)
            else:
                lower_bound = min_model_count

            if p_neg is not None:
                upper_bound = min(int(floor((range_size(p_neg) * lg) ** (1 / self.q))), max_model_count)
            else:
                upper_bound = max_model_count

            upper_bound = max(lower_bound, upper_bound)

            return ProbabilisticInterval(
                lower_bound=lower_bound,
                upper_bound=upper_bound,
                confidence=probability_of_correctness(error_probabilities),
            )

        def majority_vote_estimate(c: List[int]):
            while True:
                rf_bmc_task = make_rf_bmc_task(self._make_eamp_params(c))

                # copies the required results data in order for it not to be modified while using them
                rf_bmc_results: Counter[RfBmcResult] = \
                    self.store.data.rf_bmc_results_map.get(rf_bmc_task, Counter[RfBmcResult]()).copy()

                positive_voters = sum([
                    count
                    for result, count in rf_bmc_results.items()
                    if result.bmc is None
                ])

                negative_voters = sum([
                    count
                    for result, count in rf_bmc_results.items()
                    if result.bmc is not None
                ])

                remaining = max(0, r - (positive_voters + negative_voters))

                if positive_voters >= negative_voters and positive_voters >= negative_voters + remaining:
                    return True, majority_vote_error_probability(Fraction(1, 4), r)

                if negative_voters > positive_voters and negative_voters > positive_voters + remaining:
                    return False, majority_vote_error_probability(Fraction(1, 4), r)

                yield EampEdgeSchedulerBase.AlgorithmYield(
                    required_tasks=Counter[RfBmcTask](remaining * [rf_bmc_task]),
                    predicted_required_tasks=Counter[RfBmcTask](),
                    intermediate_result=get_edge_interval(),
                )

        p = self._make_initial_partial_eamp_params()
        mv_estimate_count = 0

        while True:
            while pre_estimate(p) is False:
                next_p = self._advance_partial_eamp_params(p, False)

                if next_p is None:
                    break
                else:
                    p = next_p

            if pre_estimate(p) is False and self._advance_partial_eamp_params(p, False) is None:
                break

            if mv_estimate_count + 1 > mv_estimate_count_upper_bound:
                raise RuntimeError(
                    "Estimate iteration upper bound was incorrect. "
                    "This error is caused by an incorrect implementation "
                    "of _get_upper_bound_on_estimate_iteration_count."
                )
            else:
                mv_estimate_count += 1

            mv_estimate, mv_error_prob = yield from majority_vote_estimate(p)
            error_probabilities.append(mv_error_prob)

            if mv_estimate:
                p_pos = p

                next_p = self._advance_partial_eamp_params(p, True)

                if next_p is None:
                    raise RuntimeError(
                        "Iteration procedure cannot be terminated on a True estimate result. "
                        "This error is caused by an incorrect implementation of _advance_partial_eamp_params."
                    )
                else:
                    p = next_p
            else:
                p_neg = p

                next_p = self._advance_partial_eamp_params(p, False)

                if next_p is None:
                    break
                else:
                    p = next_p

        if p_pos is None:
            s = self._get_required_minimal_min_model_count_when_no_lower_bound_could_be_established()

            if min_model_count < s:
                bmc_task_result: Optional[Tuple[BmcTask, BmcResult]] = self.store.data.bmc_task_result

                while bmc_task_result is None or bmc_task_result[0].a < s:
                    yield EampEdgeSchedulerBase.AlgorithmYield(
                        required_tasks=Counter[BmcTask]([BmcTask(a=s)]),
                        predicted_required_tasks=Counter[BmcTask](),
                        intermediate_result=get_edge_interval(),
                    )

                    bmc_task_result = self.store.data.bmc_task_result

                if bmc_task_result[1].bmc is not None and bmc_task_result[1].bmc < s:
                    return ProbabilisticInterval(
                        lower_bound=bmc_task_result[1].bmc,
                        upper_bound=bmc_task_result[1].bmc,
                        confidence=Fraction(1),
                    )
                else:
                    min_model_count = s

        return get_edge_interval()

    def _run_algorithm(self):
        yield from self._run_algorithm_once()
        # second iteration ensures updated results are used
        return (yield from self._run_algorithm_once())

    @staticmethod
    def get_g_and_lg(a: int) -> Tuple[float, float]:
        """
        Returns the internal parameters g and G for the given a.
        """

        return (sqrt(a + 1) - 1) ** 2, (sqrt(a + 1) + 1) ** 2



import random
from enum import Enum, unique
from math import prod, log2, ceil
from typing import NamedTuple, Tuple, Any, List
from rfb_mc.restrictive_formula_module import RestrictiveFormulaModuleBase
from rfb_mc.types import Params


@unique
class EampTransformMethod(Enum):
    SORTED_ROLLING_WINDOW = "SRW"


EampParams = NamedTuple("EampParams", [
    ("c", Tuple[int, ...]),
    ("p", Tuple[int, ...]),
    ("transform_method", EampTransformMethod),
])

EampParamProperties = NamedTuple("EampParamProperties", [
    ("range_size", int),
])

EampInstanceParams = NamedTuple("EampInstanceParams", [
    ("params", EampParams),
    ("coefficients", Tuple[Tuple[Tuple[Tuple[int, ...], int, int], ...], ...]),
    ("p", Tuple[int, ...]),
])


class EampRfm(RestrictiveFormulaModuleBase[EampParams, EampParamProperties, EampInstanceParams]):
    @classmethod
    def get_guid(cls):
        return "eamp-rfm"

    @classmethod
    def encode_restrictive_formula_params(
        cls,
        params: EampParams,
    ) -> Any:
        return (
            params.c,
            params.p,
            params.transform_method.value
        )

    @classmethod
    def decode_restrictive_formula_params(
        cls,
        params: Any,
    ) -> EampParams:
        c, p, transform_method = params

        return EampParams(
            c=c,
            p=p,
            transform_method=EampTransformMethod(transform_method)
        )

    @classmethod
    def get_restrictive_formula_properties(
        cls,
        params: Params,
        restrictive_formula_params: EampParams,
    ) -> EampParamProperties:
        return EampParamProperties(
            range_size=get_range_size(restrictive_formula_params.c, restrictive_formula_params.p)
        )

    @classmethod
    def generate_restrictive_formula_instance_params(
        cls,
        params: Params,
        restrictive_formula_params: EampParams,
        q: int,
    ) -> EampInstanceParams:
        variables: List[int] = []

        for size in sorted(params.bit_width_counter.keys()):
            # add amount of variables with size q-times as they are cloned q-times
            variables += [size] * params.bit_width_counter[size] * q

        def get_slice_count_sorted_rolling_window(domain_bit_count: int) -> int:
            slice_count = 0

            queue = sorted(variables)

            while len(queue) > 0:
                x = queue.pop(0)

                if x >= domain_bit_count:
                    for i in range(x // domain_bit_count):
                        slice_count += 1

                    if (x // domain_bit_count) * domain_bit_count != x:
                        slice_count += 1
                else:
                    slice_item = [x]

                    while len(queue) > 0 and sum([y for y in slice_item]) + queue[0] <= domain_bit_count:
                        slice_item.append(queue.pop(0))

                    slice_count += 1

            return slice_count

        def get_slice_count(domain_bit_count: int) -> int:
            if restrictive_formula_params.transform_method == EampTransformMethod.SORTED_ROLLING_WINDOW:
                return get_slice_count_sorted_rolling_window(domain_bit_count)
            else:
                raise RuntimeError(f"Not implemented transform method {restrictive_formula_params.transform_method}")

        def generate_coefficients(j: int) -> Tuple[Tuple[int, ...], int, int]:
            pj = restrictive_formula_params.p[j]

            return (
                tuple([
                    random.randint(0, pj - 1) for _ in range(
                        get_slice_count(
                            int(ceil(log2(pj)))
                        )
                    )
                ]),
                random.randint(0, pj - 1),
                random.randint(0, pj - 1),
            )

        return EampInstanceParams(
            params=restrictive_formula_params,
            coefficients=tuple([
                tuple([
                    generate_coefficients(j) for _ in range(restrictive_formula_params.c[j])
                ]) for j in range(len(restrictive_formula_params.c))
            ]),
            p=restrictive_formula_params.p,
        )


def get_range_size(c: Tuple[int, ...], p: Tuple[int, ...]) -> int:
    """
    Returns the size of the range of the hash family for
    the given c and p parameters.
    """

    return prod([p[i] ** c[i] for i in range(len(c))])



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Dict, Any, Type
from rfb_mc.restrictive_formula_module import register_restrictive_formula_module
from rfb_mc.restrictive_formula_module_implementation import RestrictiveFormulaModuleImplementationBase, \
    RestrictiveFormulaInstance
from rfb_mc.types import Params, RfBmcTask, RfBmcResult, BmcResult, BmcTask

FormulaParams = TypeVar("FormulaParams")

RestrictiveFormulaInstanceGenerationArgs = TypeVar("RestrictiveFormulaInstanceGenerationArgs")


class RunnerBase(ABC, Generic[FormulaParams, RestrictiveFormulaInstanceGenerationArgs, RestrictiveFormulaInstance]):
    def __init__(self, params: Params, formula_params: FormulaParams):
        self.check_params_and_formula_params_compatibility(params, formula_params)
        self.params: Params = params
        self.formula_params: FormulaParams = formula_params

    restrictive_formula_module_implementation_map: Dict[str, Type[RestrictiveFormulaModuleImplementationBase]] = {}
    """
    Map from restrictive formula module uid to implementation class.
    """

    @classmethod
    def register_restrictive_formula_module_implementation(cls, rfmi: Type[RestrictiveFormulaModuleImplementationBase]):
        register_restrictive_formula_module(
            rfmi.get_restrictive_formula_module(),
        )

        cls.restrictive_formula_module_implementation_map[rfmi.get_restrictive_formula_module().get_guid()] = rfmi

    @classmethod
    @abstractmethod
    def check_params_and_formula_params_compatibility(cls, params: Params, formula_params: FormulaParams):
        """
        Raises an error if the params and formula_params are not compatible.
        """

        raise NotImplementedError()

    @abstractmethod
    def get_restrictive_formula_instance_generation_args(self, q: int) -> RestrictiveFormulaInstanceGenerationArgs:
        """
        Returns additional arguments required for generating the restrictive formula instance from the params.
        """

        raise NotImplementedError()

    def generate_restrictive_formula_instance(
        self, rfm_uid: str, rfm_formula_params: Any, q: int,
    ) -> RestrictiveFormulaInstance:
        imp_map = self.restrictive_formula_module_implementation_map

        if rfm_uid not in imp_map:
            raise RuntimeError(f"Restrictive Formula Module \"{rfm_uid}\" is not implemented")

        rfmi = imp_map[rfm_uid]
        rfm = rfmi.get_restrictive_formula_module()

        instance_params = rfm.generate_restrictive_formula_instance_params(
            self.params, rfm_formula_params, q,
        )
        instance_args = self.get_restrictive_formula_instance_generation_args(q)

        return rfmi.generate_restrictive_formula(self.params, instance_params, instance_args)

    @abstractmethod
    def rf_bmc(self, task: RfBmcTask) -> RfBmcResult:
        """
        Performs bounded model counting on the formula resulting from
        first replicating the original formula q-times and
        then introducing a restrictive formula condition.
        """

        raise NotImplementedError()

    @abstractmethod
    def bmc(self, task: BmcTask) -> BmcResult:
        """
        Performs bounded model counting.
        """

        raise NotImplementedError()



from functools import lru_cache
from math import sqrt, prod, log2, ceil, floor, log
from typing import Tuple, Optional, List
from rfb_mc.component.eamp.eamp_edge_scheduler_base import EampEdgeSchedulerBase
from rfb_mc.component.eamp.primes import get_lowest_prime_above_or_equal_power_of_power_of_two
from rfb_mc.component.eamp.eamp_rfm import EampParams, EampTransformMethod


class EampEdgeScheduler(EampEdgeSchedulerBase[Tuple[int, List[int]]]):
    @property
    @lru_cache(1)
    def _cn(self):
        return int(
            floor(log2(log2(self.max_model_count ** self.q / self.lg) + 1) + 1)
        ) if self.max_model_count ** self.q / self.lg >= 1 else 1

    @property
    @lru_cache(1)
    def _p(self):
        return tuple([
            get_lowest_prime_above_or_equal_power_of_power_of_two(j)
            for j in range(self._cn)
        ])

    @lru_cache(1)
    def _get_upper_bound_on_estimate_iteration_count(self) -> int:
        # maximum amount of values that need to be iterated for c[0]
        max_c0 = int(ceil(max([
            log2(self._p[i] / prod([self._p[j] for j in range(1, i)]))
            for i in range(1, self._cn)
        ]))) - 1 if self._cn > 1 else 1

        # maximum amount of required estimate iterations
        return self._cn - 1 + max_c0

    @lru_cache(1)
    def _get_required_minimal_min_model_count_when_no_lower_bound_could_be_established(self):
        return int(ceil(self.g ** (1 / self.q)))

    def _make_eamp_params(self, partial_eamp_params: Tuple[int, List[int]]) -> EampParams:
        j, c = partial_eamp_params

        return EampParams(
            p=self._p,
            c=tuple(c),
            transform_method=EampTransformMethod.SORTED_ROLLING_WINDOW,
        )

    def _make_initial_partial_eamp_params(self) -> Tuple[int, List[int]]:
        return self._cn - 1, [0] * (self._cn - 1) + [1]

    def _advance_partial_eamp_params(
        self,
        partial_eamp_params: Tuple[int, List[int]],
        estimate_result: bool
    ) -> Optional[Tuple[int, List[int]]]:
        j, c = partial_eamp_params
        c_next = c.copy()

        if estimate_result is True:
            if j == 0:
                c_next[0] += 1
                return 0, c_next
            else:
                c_next[j - 1] = 1
                return j - 1, c_next
        else:
            if j == 0:
                return None
            else:
                c_next[j] = 0
                c_next[j - 1] = 1
                return j - 1, c_next

    @staticmethod
    def get_q_for_fixed_a_that_ensures_upper_bound_for_multiplicative_gap_of_result(
        a: int,
        epsilon: float,
    ) -> int:
        """
        Returns the minimal parameter q that ensures that for the given a we have,
        get_upper_bound_for_multiplicative_gap_of_result(a, q) <= (1 + epsilon) ** 2.
        That condition is equivalent to the statement that the geometric mean of the final edge interval
        is a multiplicative approximation with error epsilon i.e.
        model_count / (1 + epsilon) <= geometric_mean <= model_count * (1 + epsilon).
        """

        g, lg = EampEdgeScheduler.get_g_and_lg(a)
        return int(ceil(0.5 * log(2 * lg / g, 1 + epsilon)))

    @staticmethod
    def get_a_for_fixed_q_that_ensures_upper_bound_for_multiplicative_gap_of_result(
        q: int,
        epsilon: float,
    ) -> int:
        """
        Returns the minimal parameter a that ensures that for the given q we have,
        get_upper_bound_for_multiplicative_gap_of_result(a, q) <= (1 + epsilon) ** 2.
        That condition is equivalent to the statement that the geometric mean of the final edge interval
        is a multiplicative approximation with error epsilon i.e.
        model_count / (1 + epsilon) <= geometric_mean <= model_count * (1 + epsilon).
        """

        if 2 ** (1 / q) >= (1 + epsilon) ** 2:
            raise ValueError(f"For epsilon={epsilon} and q={q} "
                             f"i.e. (1 + epsilon) ** 2 = {(1 + epsilon) ** 2}, higher a "
                             f"values will only be able to converge to {2 ** (1 / q)} thus epsilon "
                             f"{sqrt(2 ** (1 / q)) - 1}")

        # TODO: replace by proper formula
        a = 1
        while EampEdgeScheduler.get_upper_bound_for_multiplicative_gap_of_result(a, q) > (1 + epsilon) ** 2:
            a += 1

        return a

    @staticmethod
    def get_upper_bound_for_multiplicative_gap_of_result(a: int, q: int) -> float:
        """
        Returns an upper bound on the multiplicative gap of the final edge interval returned
        by the eamp edge scheduler.
        """

        g, lg = EampEdgeScheduler.get_g_and_lg(a)
        return (2 * lg / g) ** (1 / q)



from typing import Generic
from rfb_mc.component.multi_processing_integrator import MultiProcessingIntegratorBase, \
    IntermediateResult, Result
from rfb_mc.component.runner_z3 import RunnerZ3, FormulaParamsZ3, \
    serialize_formula_params_z3, deserialize_formula_params_z3, SerializedFormulaParamsZ3


class MultiProcessingIntegratorZ3(
    Generic[IntermediateResult, Result],
    MultiProcessingIntegratorBase[IntermediateResult, Result, FormulaParamsZ3, SerializedFormulaParamsZ3]
):
    @classmethod
    def get_runner_class(cls):
        return RunnerZ3

    @classmethod
    def deserialize_formula_params(cls, serialized_formula_params: SerializedFormulaParamsZ3) -> FormulaParamsZ3:
        return deserialize_formula_params_z3(serialized_formula_params)

    @classmethod
    def serialize_formula_params(cls, formula_params: FormulaParamsZ3) -> SerializedFormulaParamsZ3:
        return serialize_formula_params_z3(formula_params)



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Type, Any
from rfb_mc.restrictive_formula_module import RestrictiveFormulaInstanceParams, RestrictiveFormulaModuleBase
from rfb_mc.types import Params

RestrictiveFormulaInstance = TypeVar("RestrictiveFormulaInstance")

RestrictiveFormulaInstanceGenerationArgs = TypeVar("RestrictiveFormulaInstanceGenerationArgs")


class RestrictiveFormulaModuleImplementationBase(
    ABC,
    Generic[RestrictiveFormulaInstanceParams, RestrictiveFormulaInstanceGenerationArgs, RestrictiveFormulaInstance],
):
    @classmethod
    @abstractmethod
    def get_restrictive_formula_module(
        cls,
    ) -> Type[RestrictiveFormulaModuleBase[Any, Any, RestrictiveFormulaInstanceParams]]:
        """
        Restrictive formula module this class implements.
        """

        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def generate_restrictive_formula(
        cls,
        params: Params,
        instance_params: RestrictiveFormulaInstanceParams,
        args: RestrictiveFormulaInstanceGenerationArgs,
    ) -> RestrictiveFormulaInstance:
        """
        Implements the restrictive formula that is given by the instance params for the restrictive formula module.
        """

        raise NotImplementedError()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Counter, Generic, TypeVar, Generator, Union
from rfb_mc.store import StoreBase
from rfb_mc.types import RfBmcTask, BmcTask

IntermediateResult = TypeVar("IntermediateResult")

Result = TypeVar("Result")


class SchedulerBase(ABC, Generic[IntermediateResult, Result]):
    @dataclass
    class AlgorithmYield:
        required_tasks: Counter[Union[BmcTask, RfBmcTask]]
        predicted_required_tasks: Counter[Union[BmcTask, RfBmcTask]]
        intermediate_result: IntermediateResult

    def __init__(self, store: StoreBase):
        self.store = store

    @abstractmethod
    def _run_algorithm(self) -> Generator[AlgorithmYield, None, Result]:
        """
        Generator function that yields algorithm step results and will
        return the desired result. It is expected to be deterministic and
        only use information from the store without running anything itself.

        Each generator instance should be completely independent of one another.
        """

        raise NotImplementedError()

    def run(self) -> Generator[AlgorithmYield, None, Result]:
        """
        Runs the scheduler algorithm and yields intermediate results and tasks required for
        continuing the algorithm. Note that the required tasks are only relevant to an individual step i.e.
        executing all of them will not guarantee the algorithm can complete. But continuously running the required
        tasks will mean that it will eventually complete. At which point it will return the result.
        """

        return self._run_algorithm()



from typing import Generic
from rfb_mc.component.direct_integrator import DirectIntegratorBase, IntermediateResult, Result
from rfb_mc.component.runner_z3 import RunnerZ3, FormulaParamsZ3


class DirectIntegratorZ3(
    Generic[IntermediateResult, Result],
    DirectIntegratorBase[IntermediateResult, Result, FormulaParamsZ3]
):
    @classmethod
    def get_runner_class(cls):
        return RunnerZ3



from fractions import Fraction
from typing import NamedTuple

ProbabilisticInterval = NamedTuple("ProbabilisticInterval", [
    ("lower_bound", int),
    ("upper_bound", int),
    ("confidence", Fraction),
])



from typing import Dict, Tuple
from functools import lru_cache
import os
import random


lowest_prime_above_or_equal_power_of_power_of_two_dic_file_name = os.path.join(
    os.path.dirname(__file__),
    "lowest_prime_above_or_equal_power_of_power_of_two_dict.txt"
)


@lru_cache(1)
def read_lowest_prime_above_or_equal_power_of_power_of_two_dict() -> Dict[int, int]:
    def parse_line(line: str) -> Tuple[int, int]:
        ns, ps = line.split(" ")
        return int(ns, 10), int(ps, 10)

    with open(lowest_prime_above_or_equal_power_of_power_of_two_dic_file_name, "r") as f:
        lines = f.readlines()

    return {
        n: p for n, p in map(parse_line, lines)
    }


def get_lowest_prime_above_or_equal_power_of_power_of_two(j: int) -> int:
    """
    Returns the smallest prime that above or equal 2 ** (2 ** j)
    """

    primes_dict = read_lowest_prime_above_or_equal_power_of_power_of_two_dict()

    if j in primes_dict:
        return primes_dict[j]
    else:
        raise ValueError(f"No stored primes for the lowest prime above or equal 2 ** (2 ** {j}),"
                         f" the highest available is for 2 ** (2 ** {max(primes_dict.keys())})")


def miller_rabin(n: int, k: int = 40) -> bool:
    # Implementation uses the Miller-Rabin Primality Test
    # The optimal number of rounds for this test is 40
    # See http://stackoverflow.com/questions/6325576
    #     /how-many-iterations-of-rabin-miller-should-i-use-for-cryptographic-safe-primes
    # for justification

    # If number is even, it's a composite number

    if n == 2 or n == 3:
        return True

    if n % 2 == 0:
        return False

    r, s = 0, n - 1
    while s % 2 == 0:
        r += 1
        s //= 2

    for _ in range(k):
        a = random.randrange(2, n - 1)
        x = pow(a, s, n)
        if x == 1 or x == n - 1:
            continue
        for _ in range(r - 1):
            x = pow(x, 2, n)
            if x == n - 1:
                break
        else:
            return False
    return True


def get_closest_prime(x: int, only_larger: bool = False) -> int:
    """
    Generates the prime nearest to the given "x" (including x itself).
    Unless only_larger is True in which case the lowest prime above or equal x is generated.
    There is a small probability that the generated prime is not a prime.
    """

    if miller_rabin(x, k=40):
        return x

    k = 1

    while True:
        if not only_larger and x - k > 1 and miller_rabin(x - k, k=40):
            return x - k
        elif miller_rabin(x + k, k=40):
            return x + k
        else:
            k += 1



from typing import Iterable
from fractions import Fraction
from math import ceil, comb, prod


def probability_of_correctness(
    error_probabilities: Iterable[Fraction],
) -> Fraction:
    """
    Takes in the error probabilities of independent probabilistic execution.
    Returns the probability of an error occurring in any of them.
    """

    return Fraction(prod([1 - err_prob for err_prob in error_probabilities]))


def majority_vote_error_probability(
    alpha: Fraction, r: int,
) -> Fraction:
    return sum([
        comb(r, rj) * (alpha ** r) * ((1 - alpha) ** (r - rj))
        for rj in range(ceil(r / 2), r + 1)
    ])


def multi_majority_vote_iteration_count_to_ensure_beta(
    alpha: Fraction, beta: Fraction, max_majority_voting_countings: int,
) -> int:
    """
    Returns an amount of iterations for each majority vote counting such that
    the error probability of any of the majority vote counting to fail is below beta.
    :param alpha: Error probability of the procedures in the majority vote counting
    :param beta: Desired upper bound on error probability of any majority vote counting failing
    :param max_majority_voting_countings: Maximum amount of expected majority vote counting procedures
    """

    r = 1
    while (1 - majority_vote_error_probability(alpha, r))**max_majority_voting_countings < 1 - beta:
        r += 1

    return r



from abc import abstractmethod
from datetime import datetime
from time import perf_counter, sleep
from typing import Generic, Iterable, Type, Any, Tuple
from rfb_mc.runner import FormulaParams, RunnerBase
from rfb_mc.integrator import IntegratorBase, IntermediateResult, Result
from rfb_mc.scheduler import SchedulerBase
from threading import Thread

from rfb_mc.types import RfBmcTask, BmcTask, BmcResult, RfBmcResult


class DirectIntegratorBase(
    Generic[IntermediateResult, Result, FormulaParams],
    IntegratorBase[IntermediateResult, Result]
):
    """
    Class that implements instantiating runners directly.

    Class is abstract since the runner that is used must be specified.
    """

    # whether the integrator should print debug information
    PRINT_DEBUG: bool = True

    @classmethod
    def _print_debug(cls, *messages: Iterable[str]):
        """ Timestamped version of print that only prints if PRINT_DEBUG is True """
        if cls.PRINT_DEBUG:
            for message in messages:
                print(f"[{datetime.now().strftime('%H:%M:%S:%f')}] {message}")

    @classmethod
    @abstractmethod
    def get_runner_class(cls) -> Type[RunnerBase[FormulaParams, Any, Any]]:
        """
        Returns class used for the runner in worker processes.
        """
        raise NotImplementedError()

    def __init__(self, formula_params: FormulaParams):
        super().__init__(formula_params)

    def run(self, scheduler: SchedulerBase):
        runner = self.get_runner_class()(
            params=scheduler.store.data.params,
            formula_params=self.formula_params,
        )

        algorithm_generator = scheduler.run()
        prev_intermediate_result = None

        s1 = perf_counter()

        self._print_debug("Starting integrator run")

        try:
            # execute tasks until the algorithm stops the iteration thus indicating the final result
            while True:
                # execute an algorithm step
                algorithm_yield = next(algorithm_generator)

                # if the intermediate result has changed, it should be published via a yield
                if algorithm_yield.intermediate_result != prev_intermediate_result:
                    prev_intermediate_result = algorithm_yield.intermediate_result
                    yield prev_intermediate_result

                if sum(algorithm_yield.required_tasks.values()) > 0:
                    task = next(algorithm_yield.required_tasks.elements())

                    s = perf_counter()

                    if type(task) == BmcTask:
                        task_result: Tuple[BmcTask, BmcResult] = (task, runner.bmc(task))
                    else:
                        task_result: Tuple[RfBmcTask, RfBmcResult] = (task, runner.rf_bmc(task))

                    self._print_debug(f"Ran {task_result[0]} returning {task_result[1]}"
                                      f" which took {perf_counter() - s:.3f} seconds")

                    # starts result synchronization with store asynchronously
                    Thread(
                        target=scheduler.store.add_results,
                        kwargs={
                          "task_results": [task_result],
                        },
                        daemon=True,
                    ).start()

                    # switch thread to allow adding results to store before executing next algorithm iteration,
                    # thus the synchronous part of the add_rf_bmc_results function will execute before the algorithm
                    # is continued
                    sleep(0)
        except StopIteration as err:
            d1 = perf_counter() - s1
            self._print_debug(f"Running scheduler tasks until result was available took {d1:.2f} seconds")
            self._print_debug(f"Result: {err.value}")

            return err.value



from typing import Iterable, Tuple
from rfb_mc.store import StoreBase
from rfb_mc.types import RfBmcTask, RfBmcResult, BmcTask, BmcResult


class InMemoryStore(StoreBase):
    """
    Only stores in memory
    """

    def sync(self):
        pass

    def _add_results(
        self,
        bmc_task_results: Iterable[Tuple[BmcTask, BmcResult]],
        rf_bmc_task_results: Iterable[Tuple[RfBmcTask, RfBmcResult]],
    ):
        pass



from abc import abstractmethod
from datetime import datetime
from multiprocessing import Lock, Queue, Process
from queue import Empty
from time import perf_counter, sleep
import os
from collections import Counter
from threading import Thread
from typing import Generic, Iterable, Type, TypeVar, Any
from rfb_mc.runner import FormulaParams, RunnerBase
from rfb_mc.integrator import IntegratorBase, IntermediateResult, Result
from rfb_mc.scheduler import SchedulerBase
from rfb_mc.types import Params, RfBmcTask, BmcTask

SerializedFormulaParams = TypeVar("SerializedFormulaParams")


class MultiProcessingIntegratorBase(
    Generic[IntermediateResult, Result, FormulaParams, SerializedFormulaParams],
    IntegratorBase[IntermediateResult, Result]
):
    """
    Class that implements instantiating runners in created processes thus enabling parallel
    execution of scheduler tasks on multiple CPU cores.

    Class is abstract since the runner that is used must be specified and it must be specified how the formula params
    are serialized to enable them being transferred using the python multiprocessing process arguments.
    """

    # whether the integrator should print debug information
    PRINT_DEBUG: bool = True

    @classmethod
    def _print_debug(cls, *messages: Iterable[str]):
        """ Timestamped version of print that only prints if PRINT_DEBUG is True """
        if cls.PRINT_DEBUG:
            for message in messages:
                print(f"[{datetime.now().strftime('%H:%M:%S:%f')}] {message}")

    @classmethod
    @abstractmethod
    def get_runner_class(cls) -> Type[RunnerBase[FormulaParams, Any, Any]]:
        """
        Returns class used for the runner in worker processes.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def serialize_formula_params(cls, formula_params: FormulaParams) -> SerializedFormulaParams:
        """
        Returns a serialized version of the formula params that can be transferred
        via multiprocessing process arguments.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def deserialize_formula_params(cls, serialized_formula_params: SerializedFormulaParams) -> FormulaParams:
        """
        Returns formula params from the serialized formula params.
        """
        raise NotImplementedError()

    @classmethod
    def _run_worker(
        cls,
        worker_idx: int,
        worker_count: int,
        stdio_lock: Lock,
        task_queue: Queue,
        result_queue: Queue,
        params: Params,
        serialized_formula_params: FormulaParams,
    ):
        worker_number = worker_idx + 1

        def print_debug(*messages: Iterable[str]):
            """
            Print debug version that ensures only one source is writing to the standard input output at a time
            """

            # PRINT_DEBUG is checked because even tough it is already enforced in _print_debug,
            # using it here will prevent the lock from being acquired as it would waste time if
            # printing is not used
            if cls.PRINT_DEBUG:
                with stdio_lock:
                    cls._print_debug(
                        *[f"Worker[{worker_number}/{worker_count}]: {message}" for message in messages],
                    )

        runner = cls.get_runner_class()(
            params=params,
            formula_params=cls.deserialize_formula_params(serialized_formula_params),
        )

        print_debug("Initialized")

        # retrieve and execute tasks, until the queued task is None,
        # which the integrator puts if it has finished

        task = task_queue.get()

        while task is not None:
            s = perf_counter()
            result = runner.bmc(task) if type(task) == BmcTask else runner.rf_bmc(task)
            result_queue.put((task, result))
            print_debug(f"Ran {task} returning {result} which took {perf_counter() - s:.3f} seconds")

            task = task_queue.get()

    def __init__(self, formula_params: FormulaParams, worker_count: int = os.cpu_count()):
        super().__init__(formula_params)
        self.worker_count = worker_count

    def run(self, scheduler: SchedulerBase):
        task_queue = Queue()
        result_queue = Queue()
        stdio_lock = Lock()

        def print_debug(*messages: Iterable[str]):
            # same as for the workers
            if self.PRINT_DEBUG:
                with stdio_lock:
                    self._print_debug(
                        *[f"Integrator: {message}" for message in messages]
                    )

        print_debug("Starting integrator run")

        processes = [
            Process(
                target=self._run_worker,
                kwargs={
                    "worker_idx": worker_idx,
                    "worker_count": self.worker_count,
                    "stdio_lock": stdio_lock,
                    "task_queue": task_queue,
                    "result_queue": result_queue,
                    "params": scheduler.store.data.params,
                    "serialized_formula_params": self.serialize_formula_params(self.formula_params),
                },
                daemon=True,
            ) for worker_idx in range(self.worker_count)
        ]

        for process in processes:
            process.start()

        s1 = perf_counter()

        try:
            tasks_in_progress: Counter[RfBmcTask] = Counter()

            algorithm_generator = scheduler.run()
            prev_intermediate_result = None

            try:
                # execute tasks until the algorithm stops the iteration thus indicating the final result
                while True:
                    # execute an algorithm step
                    algorithm_yield = next(algorithm_generator)

                    # if the intermediate result has changed, it should be published via a yield
                    if algorithm_yield.intermediate_result != prev_intermediate_result:
                        prev_intermediate_result = algorithm_yield.intermediate_result

                        print_debug(f"Intermediate Result: {prev_intermediate_result}")

                        yield prev_intermediate_result

                    # determine how many tasks should be queued, by first considering what tasks
                    # are not yet in progress and how many idle workers exist
                    required_tasks = algorithm_yield.required_tasks - tasks_in_progress
                    idle_workers = self.worker_count - sum(tasks_in_progress.values())
                    tasks_to_queue = min(sum(required_tasks.values()), idle_workers)

                    # TODO: consider queueing predicted tasks if workers are idle and no additional required tasks exist

                    # queue as many tasks as are available and can be directly forwarded to an idle worker
                    if tasks_to_queue > 0:
                        for _ in range(tasks_to_queue):
                            task = [task for task, count in required_tasks.items() if count > 0][0]

                            required_tasks -= Counter([task])
                            tasks_in_progress += Counter([task])
                            task_queue.put(task)

                    # if tasks are in progress we wait until a result is available, since either all workers are
                    # processing or no more tasks are available until existing tasks have been finished
                    if sum(tasks_in_progress.values()) > 0:
                        # waits for at least one result and if multiple are available, retrieves all
                        task_results = []
                        task_result = result_queue.get()
                        while task_result is not None:
                            task_results.append(task_result)

                            try:
                                task_result = result_queue.get_nowait()
                            except Empty:
                                task_result = None

                        # add all results to the store
                        Thread(
                            target=scheduler.store.add_results,
                            kwargs={
                                "task_results": task_results,
                            },
                        ).start()

                        # switch thread to allow adding results to store before executing next algorithm iteration,
                        # thus the synchronous part of the add_rf_bmc_results function will execute before the algorithm
                        # is continued
                        sleep(0)

                        # remove accomplished tasks from in progress counter
                        for task, result in task_results:
                            tasks_in_progress -= Counter([task])
            except StopIteration as err:
                d1 = perf_counter() - s1
                print_debug(f"Running schedulers tasks until result was available took {d1:.2f} seconds")
                print_debug(f"Result: {err.value}")

                return err.value
        finally:
            for process in processes:
                process.terminate()



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Hashable, Any, Dict, Type
from rfb_mc.types import Params

RestrictiveFormulaParams = TypeVar("RestrictiveFormulaParams", bound=Hashable)
# parameter that determines all formula generation related values

RestrictiveFormulaProperties = TypeVar("RestrictiveFormulaProperties")
# properties of the restrictive formula that are determined by the parameters

RestrictiveFormulaInstanceParams = TypeVar("RestrictiveFormulaInstanceParams")


# values that parameterize a specific instance generated by the module and will be used by
# the runner implementations to reconstruct the formula in the format of which ever solver is used


class RestrictiveFormulaModuleBase(
    Generic[RestrictiveFormulaParams, RestrictiveFormulaProperties, RestrictiveFormulaInstanceParams],
    ABC,
):
    @classmethod
    @abstractmethod
    def get_guid(cls) -> str:
        """
        UID of the restrictive formula module that needs to be deterministic and
        unique across all other restrictive formula module implementations.

        An abbreviation of the name that is unlikely to be reused is suggested and possibly a
        version number in order to differentiate between different generations of the same module.
        """

        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def encode_restrictive_formula_params(
        cls,
        params: Params,
    ) -> Any:
        """
        Encodes the formula parameters into a native python type.
        """

        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def decode_restrictive_formula_params(
        cls,
        params: Any,
    ) -> Params:
        """
        Decodes the formula parameters from the native python type generated by the encoder.
        """

        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def get_restrictive_formula_properties(
        cls,
        params: Params,
        restrictive_formula_params: RestrictiveFormulaParams,
    ) -> RestrictiveFormulaProperties:
        """
        Returns properties that the restrictive formula generation posses with the given parameters.
        """

        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def generate_restrictive_formula_instance_params(
        cls,
        params: Params,
        restrictive_formula_params: RestrictiveFormulaParams,
        q: int,
    ) -> RestrictiveFormulaInstanceParams:
        """
        Generate the restrictive formula instance params for the given params using randomness from the random runner
        class.
        """

        raise NotImplementedError()


restrictive_formula_module_map: Dict[str, Type[RestrictiveFormulaModuleBase]] = {}
"""
Map from restrictive formula module uid to module class.
"""


def register_restrictive_formula_module(rfm: Type[RestrictiveFormulaModuleBase]):
    restrictive_formula_module_map[rfm.get_guid()] = rfm


def get_restrictive_formula_module(uid: str) -> Type[RestrictiveFormulaModuleBase]:
    if uid in restrictive_formula_module_map:
        return restrictive_formula_module_map[uid]
    else:
        raise RuntimeError(f"Restrictive formula module \"{uid}\" is required but is not registered")



from ast import literal_eval
from typing import Dict, Tuple, TypedDict, Any, Literal, Counter, Optional

from rfb_mc.restrictive_formula_module import get_restrictive_formula_module
from rfb_mc.store import StoreData
from rfb_mc.types import RfBmcTask, RfBmcResult, Params, BmcTask, BmcResult


def v1_encode_rf_bmc_task(task: RfBmcTask) -> str:
    return repr((
        task.rfm_guid,
        get_restrictive_formula_module(task.rfm_guid).encode_restrictive_formula_params(
            task.rfm_formula_params,
        ),
        task.a,
        task.q
    ))


def v1_decode_rf_bmc_task(task: str) -> RfBmcTask:
    rfm_guid, rfm_formula_params, a, q = literal_eval(task)

    return RfBmcTask(
        rfm_guid=rfm_guid,
        rfm_formula_params=get_restrictive_formula_module(rfm_guid).decode_restrictive_formula_params(
            rfm_formula_params,
        ),
        a=a,
        q=q
    )


def v1_encode_rf_bmc_result(result: RfBmcResult) -> str:
    return repr(tuple(result))


def v1_decode_rf_bmc_result(result: str) -> RfBmcResult:
    return RfBmcResult(*literal_eval(result))


def v1_encode_rf_bmc_task_result(task_result: Tuple[RfBmcTask, RfBmcResult]) -> str:
    return repr((
        v1_encode_rf_bmc_task(task_result[0]),
        v1_encode_rf_bmc_result(task_result[1])
    ))


def v1_decode_rf_bmc_task_result(task_result: str) -> Tuple[RfBmcTask, RfBmcResult]:
    task, result = literal_eval(task_result)

    return (
        v1_decode_rf_bmc_task(task),
        v1_decode_rf_bmc_result(result),
    )


def v1_encode_bmc_task(task: BmcTask) -> str:
    return repr((
        task.a,
    ))


def v1_decode_bmc_task(task: str) -> BmcTask:
    a, = literal_eval(task)

    return BmcTask(
        a=a,
    )


def v1_encode_bmc_result(result: BmcResult) -> str:
    return repr(tuple(result))


def v1_decode_bmc_result(result: str) -> BmcResult:
    return BmcResult(*literal_eval(result))


def v1_encode_bmc_task_result(task_result: Tuple[BmcTask, BmcResult]) -> str:
    return repr((
        v1_encode_bmc_task(task_result[0]),
        v1_encode_bmc_result(task_result[1])
    ))


def v1_decode_bmc_task_result(task_result: str) -> Tuple[BmcTask, BmcResult]:
    task, result = literal_eval(task_result)

    return (
        v1_decode_bmc_task(task),
        v1_decode_bmc_result(result),
    )


SerializedV1RfBmcResultsMap = Dict[str, int]


def v1_encode_rf_bmc_results_map(
    rf_bmc_results_map: Dict[RfBmcTask, Counter[RfBmcResult]],
) -> SerializedV1RfBmcResultsMap:
    return {
        v1_encode_rf_bmc_task_result((task, result)): rf_bmc_results_map[task][result]
        for task in rf_bmc_results_map
        for result in rf_bmc_results_map[task]
    }


def v1_decode_rf_bmc_results_map(
    rf_bmc_results_map: SerializedV1RfBmcResultsMap,
) -> Dict[RfBmcTask, Counter[RfBmcResult]]:
    task_results = list(map(v1_decode_rf_bmc_task_result, rf_bmc_results_map.keys()))

    tasks = set([task for task, _ in task_results])

    return {
        task: Counter[RfBmcResult]({
            result: int(rf_bmc_results_map[v1_encode_rf_bmc_task_result((task, result))])
            for result in [task_result[1] for task_result in task_results if task_result[0] == task]
        })
        for task in tasks
    }


class SerializedV1Params(TypedDict):
    bit_width_counter: Dict[str, int]


def v1_encode_params(params: Params) -> SerializedV1Params:
    return {
        "bit_width_counter": {
            str(key): params.bit_width_counter[key]
            for key in params.bit_width_counter.keys()
        }
    }


def v1_decode_params(params: SerializedV1Params) -> Params:
    return Params(
        bit_width_counter=Counter[int]({
            int(key): int(params["bit_width_counter"][key])
            for key in params["bit_width_counter"]
        })
    )


# Python representation of the store data with native types,
# i.e. processable using pickle, repr/literal_eval and json.
class SerializedV1StoreData(TypedDict):
    version: Literal[1]
    params: SerializedV1Params
    rf_bmc_results_map: SerializedV1RfBmcResultsMap
    bmc_task_result: Optional[str]


def v1_encode_store_data(data: StoreData) -> SerializedV1StoreData:
    return SerializedV1StoreData(
        version=1,
        params=v1_encode_params(data.params),
        rf_bmc_results_map=v1_encode_rf_bmc_results_map(
            data.rf_bmc_results_map,
        ),
        bmc_task_result=None if data.bmc_task_result is None else v1_encode_bmc_task_result(data.bmc_task_result),
    )


def v1_decode_store_data(data: SerializedV1StoreData) -> StoreData:
    return StoreData(
        params=v1_decode_params(data["params"]),
        rf_bmc_results_map=v1_decode_rf_bmc_results_map(data["rf_bmc_results_map"]),
        bmc_task_result=None if data["bmc_task_result"] is None else v1_decode_bmc_task_result(data["bmc_task_result"]),
    )


def decode_store_data(data: Any) -> Tuple[int, StoreData]:
    """
    Decodes a serialized store data item of any known version,
    returns the version it was encoded in and the interpreted store data
    """

    if data["version"] == 1:
        typed_item: SerializedV1StoreData = data
        return data["version"], v1_decode_store_data(typed_item)
    else:
        raise ValueError(f"Unexpected version \"{data['version']}\" in store data entry")



from decimal import Decimal
from collections import Counter
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Iterable, Tuple, TypedDict
import uuid
from rfb_mc.serialization import SerializedV1StoreData, v1_encode_rf_bmc_task_result, decode_store_data, \
    v1_encode_store_data, v1_encode_params, v1_encode_bmc_task_result
from rfb_mc.store import StoreBase, StoreData
from rfb_mc.types import RfBmcTask, RfBmcResult, Params, BmcTask, BmcResult


class DynamodbV1StoreItemBmcTask(TypedDict):
    a: Decimal


class DynamodbV1StoreItemBmcResult(TypedDict):
    bmc: Optional[Decimal]


class DynamodbV1StoreItemBmcTaskResult(TypedDict):
    task: DynamodbV1StoreItemBmcTask
    result: DynamodbV1StoreItemBmcResult


def v1_convert_dynamodb_store_item_bmc_task_result(
    task_result: DynamodbV1StoreItemBmcTaskResult,
) -> Tuple[BmcTask, BmcResult]:
    return BmcTask(
        a=int(task_result["task"]["a"]),
    ), BmcResult(
        bmc=int(task_result["result"]["bmc"]),
    )


def v1_convert_bmc_task_result(
    task_result: Tuple[BmcTask, BmcResult],
) -> DynamodbV1StoreItemBmcTaskResult:
    return DynamodbV1StoreItemBmcTaskResult(
        task=DynamodbV1StoreItemBmcTask(
            a=Decimal(task_result[0].a),
        ),
        result=DynamodbV1StoreItemBmcResult(
            bmc=Decimal(task_result[1].bmc),
        )
    )


class DynamodbV1StoreItem(SerializedV1StoreData):
    id: str
    bmc_task_result: Optional[DynamodbV1StoreItemBmcTaskResult]


def v1_convert_dynamodb_store_item(item: DynamodbV1StoreItem) -> SerializedV1StoreData:
    return SerializedV1StoreData(
        version=1,
        params=item["params"],
        rf_bmc_results_map=item["rf_bmc_results_map"],
        bmc_task_result=v1_encode_bmc_task_result(
            v1_convert_dynamodb_store_item_bmc_task_result(item["bmc_task_result"])
        ) if item["bmc_task_result"] is not None else None,
    )


def v1_convert_store_data(store_data: StoreData) -> DynamodbV1StoreItem:
    return DynamodbV1StoreItem(
        **v1_encode_store_data(store_data),
        bmc_task_result=v1_convert_bmc_task_result(store_data.bmc_task_result)
        if store_data.bmc_task_result is not None else None
    )


class DynamodbStore(StoreBase):
    VERSION = 1

    def __init__(self, table, ident: str):
        """
        Initializes a dynamodb store, requires the identifier to point to
        an existing store data entry. It modifies the data format if the version is
        different as otherwise update methods will throw.
        """

        super().__init__(
            DynamodbStore.get_and_correct_store_data_entry(table, ident)
        )

        self.table = table
        self.ident = ident

    def sync(self):
        data = self.get_store_data_entry(self.table, self.ident)[1]

        with self.data_lock:
            self.data = data

    def _add_results(
        self,
        bmc_task_result: Optional[Tuple[BmcTask, BmcResult]],
        rf_bmc_task_results: Iterable[Tuple[RfBmcTask, RfBmcResult]],
    ):
        # dynamodb request to increment rf bmc task result counters

        def send_rf_bmc_request(
            task_result: Tuple[RfBmcTask, RfBmcResult],
            count: int,
        ):
            expression_attribute_names = {
                "#task_result": v1_encode_rf_bmc_task_result(task_result),
            }

            expression_attribute_values = {
                ":version": DynamodbStore.VERSION,
                ":inc": count,
            }

            update_expression = "ADD rf_bmc_results_map.#task_result :inc"

            # increments the necessary counters
            return self.table.update_item(
                Key={"id": self.ident},
                UpdateExpression=update_expression,
                ConditionExpression="attribute_exists(id) AND version = :version",
                ExpressionAttributeValues=expression_attribute_values,
                ExpressionAttributeNames=expression_attribute_names,
            )

        # dynamodb request to update the bmc task result

        def send_bmc_request(task_result: Tuple[BmcTask, BmcResult]):
            dyn_task_result = v1_convert_bmc_task_result(task_result)

            try:
                # applies the bmc task result
                return self.table.update_item(
                    Key={"id": self.ident},
                    UpdateExpression="SET bmc_task_result = :bmc_task_result",
                    ConditionExpression="attribute_exists(id) AND version = :version AND "
                                        "(attribute_not_exists(bmc_task_result) "
                                        "OR bmc_task_result = :null "
                                        "OR bmc_task_result.task.a <= :bmc_task_a)",
                    ExpressionAttributeValues={
                        ":null": None,
                        ":version": DynamodbStore.VERSION,
                        ":bmc_task_result": dyn_task_result,
                        ":bmc_task_a": dyn_task_result["task"]["a"],
                    },
                )
            except self.table.meta.client.exceptions.ConditionalCheckFailedException:
                return None

        with ThreadPoolExecutor() as executor:
            rf_bmc_task_result_counter = Counter(rf_bmc_task_results)

            fs = []

            if bmc_task_result:
                fs.append(executor.submit(send_bmc_request, bmc_task_result))

            for task_result in rf_bmc_task_result_counter:
                fs.append(executor.submit(send_rf_bmc_request, task_result, rf_bmc_task_result_counter[task_result]))

            for fut in fs:
                fut.result()

        self.sync()

    @classmethod
    def get_and_correct_store_data_entry(
        cls,
        table,
        ident: str,
    ) -> StoreData:
        """
        Retrieves the store data and updates the data format if the version is
        different.
        """

        version, data = DynamodbStore.get_store_data_entry(table, ident)

        # ensures the data format is correct in order for class method to
        # update the data correctly
        if version != cls.VERSION:
            cls.replace_store_data_entry(table, ident, data)

        return data

    @staticmethod
    def get_store_data_entry(
        table,
        ident: str,
    ) -> (int, StoreData):
        """
        Retrieves the store data entry with the given identifier from
        the table and decodes it.
        """

        item: DynamodbV1StoreItem = table.get_item(
            Key={
                "id": ident,
            }
        )["Item"]

        return decode_store_data(v1_convert_dynamodb_store_item(item))

    @staticmethod
    def replace_store_data_entry(
        table,
        ident: str,
        store_data: StoreData,
    ):
        """
        Removes the store entry and then puts the provided data in the entry.
        """

        table.delete_item(
            Key={
                "id": ident,
            },
        )

        item: DynamodbV1StoreItem = v1_convert_store_data(store_data)

        table.put_item(
            Item=item
        )

    @staticmethod
    def create_store_data_entry(
        table,
        params: Params,
        ident: Optional[str] = None,
        accept_existing: bool = False,
    ) -> str:
        """
        Creates an empty store entry.
        If the ident is specified it will be used, otherwise a uuid4 id will be generated.
        If accept_existing is True and the ident is specified it will not raise an error if there already
        exist a store entry with the given ident.
        Note that if ident is not specified, this method will retry until an ident is generated that does not already
        exist.
        """

        ident_specified = ident is not None
        # a generated uuid4 id is highly unlikely to collide with existing ids
        ident = ident if ident_specified else str(uuid.uuid4())

        item: DynamodbV1StoreItem = DynamodbV1StoreItem(
            id=ident,
            version=1,
            params=v1_encode_params(params),
            rf_bmc_results_map={},
            bmc_task_result=None,
        )

        try:
            table.put_item(
                Item=item,
                ConditionExpression="attribute_not_exists(id)",
            )
        except table.meta.client.exceptions.ConditionalCheckFailedException:
            if ident_specified:
                if accept_existing:
                    return ident
                else:
                    raise RuntimeError(f"Store entry with ident \"{ident}\" already exists")
            else:
                # retry creating a store entry since the id was already generated before
                return DynamodbStore.create_store_data_entry(
                    table, params, None, accept_existing
                )

        return ident



from abc import ABC, abstractmethod
from typing import Dict, Counter, Iterable, Tuple, Optional, cast, Union
from dataclasses import dataclass, field
from rfb_mc.types import Params, RfBmcTask, RfBmcResult, BmcResult, BmcTask
from threading import Lock


@dataclass
class StoreData:
    # general and problem specific parameter for the hash based model counting framework
    params: Params
    # results from hashed bounded model counting calls
    rf_bmc_results_map: Dict[RfBmcTask, Counter[RfBmcResult]] = field(default_factory=dict)
    # result from the bounded model count with the highest parameter for "a"
    bmc_task_result: Optional[Tuple[BmcTask, BmcResult]] = None


class StoreBase(ABC):
    def __init__(self, data: StoreData):
        self.data = data
        self.data_lock = Lock()

    @abstractmethod
    def sync(self):
        """
        Synchronizes the memory with the storage location
        used by the store implementation.

        (Possibly causes a blocking operation)
        """

        raise NotImplementedError()

    @abstractmethod
    def _add_results(
        self,
        bmc_task_result: Optional[Tuple[BmcTask, BmcResult]],
        rf_bmc_task_results: Iterable[Tuple[RfBmcTask, RfBmcResult]],
    ):
        """
        Should implement adding the results and synchronizing the external store.
        """

        raise NotImplementedError()

    def add_results(self, task_results: Iterable[Union[Tuple[RfBmcTask, RfBmcResult], Tuple[BmcTask, BmcResult]]]):
        """
        Adds a result of a rf bmc call to the data.
        Based on the store implementation this operation should also
        synchronize with the storage location.

        (Possibly causes a blocking operation)
        """

        bmc_task_results: Iterable[Tuple[BmcTask, BmcResult]] = [
            cast(Tuple[BmcTask, BmcResult], task_result) for task_result in task_results
            if isinstance(task_result[0], BmcTask)
        ]

        rf_bmc_task_results: Iterable[Tuple[RfBmcTask, RfBmcResult]] = [
            cast(Tuple[RfBmcTask, RfBmcResult], task_result) for task_result in task_results
            if isinstance(task_result[0], RfBmcTask)
        ]

        # only bmc task result with highest "a" value can possibly be set
        bmc_task_result = max(bmc_task_results, key=lambda task_result: task_result[0].a, default=None)

        with self.data_lock:
            for task, result in rf_bmc_task_results:
                if task not in self.data.rf_bmc_results_map:
                    self.data.rf_bmc_results_map[task] = Counter[RfBmcResult]()

                self.data.rf_bmc_results_map[task][result] += 1

            if self.data.bmc_task_result is None or self.data.bmc_task_result[0].a <= bmc_task_result[0].a:
                self.data.bmc_task_result = bmc_task_result

        self._add_results(bmc_task_result, rf_bmc_task_results)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rfb_mc",
    version="0.0.17",
    author="Jonah Leander Hoff",
    author_email="jonah-hoff@outlook.com",
    description="Performs model counting using restrictive formulas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Meterius/rfb-mc",
    project_urls={
        "Bug Tracker": "https://github.com/Meterius/rfb-mc/issues",
    },
    packages=setuptools.find_packages(exclude=["rfb_mc.test.*", "rfb_mc.test"]),
    include_package_data=True,
    python_requires=">=3.6",
)


import z3
from math import log2, ceil
from typing import Type, List, Tuple
from rfb_mc.component.eamp.eamp_rfm import EampInstanceParams, EampRfm, EampTransformMethod
from rfb_mc.component.runner_z3 import RfmiGenerationArgsZ3
from rfb_mc.restrictive_formula_module_implementation import RestrictiveFormulaModuleImplementationBase
from rfb_mc.types import Params


class EampRfmiZ3(RestrictiveFormulaModuleImplementationBase[EampInstanceParams, RfmiGenerationArgsZ3, z3.BoolRef]):
    @classmethod
    def get_restrictive_formula_module(
        cls,
    ) -> Type[EampRfm]:
        return EampRfm

    @classmethod
    def generate_restrictive_formula(
        cls,
        params: Params,
        instance_params: EampInstanceParams,
        args: RfmiGenerationArgsZ3,
    ) -> z3.BoolRef:
        def get_slices_sorted_rolling_window(domain_bit_count: int) -> List[z3.BitVecRef]:
            slices = []

            queue = sorted(args.variables, key=lambda x: x.size())

            while len(queue) > 0:
                x = queue.pop(0)

                if x.size() >= domain_bit_count:
                    for i in range(x.size() // domain_bit_count):
                        slices.append(
                            z3.Extract(i * domain_bit_count + domain_bit_count - 1, i * domain_bit_count, x)
                        )

                    if (x.size() // domain_bit_count) * domain_bit_count != x.size():
                        rem_slice_size = x.size() % domain_bit_count
                        slices.append(
                            z3.ZeroExt(
                                domain_bit_count - rem_slice_size,
                                z3.Extract(x.size() - 1, x.size() - rem_slice_size, x)
                            )
                        )
                else:
                    slice_item = [x]

                    while len(queue) > 0 and sum([y.size() for y in slice_item]) + queue[0].size() <= domain_bit_count:
                        slice_item.append(queue.pop(0))

                    slices.append(z3.Concat(slice_item) if len(slice_item) > 1 else slice_item[0])

            return slices

        def get_slices(domain_bit_count: int) -> List[z3.BitVecRef]:
            if instance_params.params.transform_method == EampTransformMethod.SORTED_ROLLING_WINDOW:
                return get_slices_sorted_rolling_window(domain_bit_count)
            else:
                raise RuntimeError(f"Not implemented transform method {instance_params.params.transform_method}")

        def make_hash_equation(pj: int, var_b: Tuple[int], b1: int, b2: int):
            domain_bit_count = int(ceil(log2(pj)))
            bc = int(ceil(log2(pj + 1)))
            slices = get_slices(domain_bit_count)

            return z3.URem(
                z3.Sum([
                    z3.ZeroExt(bc - s.size(), s) * z3.BitVecVal(var_b[i], bc) for i, s in enumerate(slices)
                ]) + z3.BitVecVal(b1, bc),
                z3.BitVecVal(pj, bc)
            ) == z3.BitVecVal(b2, bc)

        return z3.And([
            make_hash_equation(instance_params.p[j], var_b, b1, b2)
            for j in range(len(instance_params.p))
            for var_b, b1, b2 in instance_params.coefficients[j]
        ])



from typing import List, Any
from copy import copy


def omit_typed_dict(typed_dict: Any, keys: List[str]):
    typed_dict_copy = copy(typed_dict)

    for key in keys:
        del typed_dict_copy[key]

    return typed_dict_copy



import z3
from typing import List, NamedTuple, Dict, Optional

CloneExpressionOutput = NamedTuple("CloneExpressionOutput", [
    ("clones", List[z3.BoolRef]), ("var_map", Dict[z3.ExprRef, List[z3.ExprRef]])
])


def serialize_expression(expression: z3.ExprRef) -> str:
    s = z3.Solver()
    s.add(expression)
    return s.sexpr()


def deserialize_expression(serialized_expression: str, ctx: Optional[z3.Context] = None) -> z3.ExprRef:
    return z3.And(z3.parse_smt2_string(serialized_expression, ctx=ctx))


def get_variables(expression: z3.ExprRef) -> List[z3.ExprRef]:
    """
    Returns all variables that are contained in the expression.
    :param expression: Expression from which variables are extracted
    """

    class AstRefKey:
        def __init__(self, n):
            self.n = n

        def __hash__(self):
            return self.n.hash()

        def __eq__(self, other):
            return self.n.eq(other.n)

        def __repr__(self):
            return str(self.n)

    def askey(n):
        assert isinstance(n, z3.AstRef)
        return AstRefKey(n)

    variables = set()

    def collect(f):
        if z3.is_const(f):
            if f.decl().kind() == z3.Z3_OP_UNINTERPRETED and not askey(f) in variables:
                variables.add(askey(f))
        else:
            for c in f.children():
                collect(c)

    collect(expression)
    return [elem.n for elem in variables]


def recreate_variable(key: str, variable: z3.ExprRef) -> z3.ExprRef:
    """
    Recreates the variable but renames it with a key that is used
    to make it distinct.
    :param key:
    :param variable:
    """

    return z3.Const(f"{key}_{variable}", variable.sort())


def clone_expression(
    expression: z3.ExprRef,
    q: int,
    required_variables: Optional[List[z3.ExprRef]] = None,
) -> CloneExpressionOutput:
    """
    Clones expression by generating q instances of the expression where each
    variable is substituted by a unique newly generated variable for each variable in each clone.
    The output will list each clone and a dictionary where each entry corresponds to
    a mapping from variable in the original formula to the substituted cloned variables for each clone
    listed in the same order as the clone list.
    :param expression: Expression to be cloned
    :param q: Amount of clones created
    :param required_variables: Variables that should be cloned and put into the var_map even if they are not
    contained in the expression.
    """

    variables = set(get_variables(expression)).union(set(required_variables or []))

    var_map = {
        x: [recreate_variable(f"clone{{{i}}}", x) for i in range(q)] for x in variables
    }

    clones = [z3.substitute(expression, [(x, var_map[x][i]) for x in variables]) for i in range(q)]

    return CloneExpressionOutput(
        clones=clones,
        var_map=var_map,
    )



from abc import abstractmethod, ABC
from typing import Generator, Generic
from rfb_mc.runner import FormulaParams
from rfb_mc.scheduler import SchedulerBase, IntermediateResult, Result


class IntegratorBase(ABC, Generic[IntermediateResult, Result]):
    def __init__(self, formula_params: FormulaParams):
        self.formula_params = formula_params

    @abstractmethod
    def run(self, scheduler: SchedulerBase[IntermediateResult, Result]) -> Generator[IntermediateResult, None, Result]:
        """
        Runs the scheduler algorithm and orchestrates runners to execute the tasks that are required for
        its completion. Thus this runs the scheduler algorithm and only returns the intermediate results and the end
        result.
        """

        raise NotImplementedError()

    def run_all(self, scheduler: SchedulerBase[IntermediateResult, Result]) -> Result:
        """
        Like run, but will discard intermediate results and thus instead of being a generator this is
        a proper function that will only return the end result.
        """

        run_generator = self.run(scheduler)

        try:
            while True:
                next(run_generator)
        except StopIteration as err:
            return err.value
